#-------------------------------------------------------------------------#
# e2mysql.awk                                                             #
#                                                                         #
# Simple ESQL precompiler to convert embedded SQL for use                 #
# with an MySQL database.                                                 #
#                                                                         #
# Michael Benzinger, Sabre, June 2002                                     #
#                                                                         #
# This script was adapted from original work done by Alan                 #
# Walker, Sabre Labs.                                                     #
#                                                                         #
# This is free software; you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation; either version 2 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# This software is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this software; if not, write to the Free Software            #
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,  #
# USA.                                                                    #
#-------------------------------------------------------------#
BEGIN {
  print "// Code precompiled by e2mysql.awk";
  print "#include \"e2mysql.h\"";

  buffer    = "";
  indentStr = "    ";
  lineCount =  0;
  lineNo    = -1;
  uniqCount =  1;
  ifCount   =  0;
  tandem    =  0;
  inDeclare =  0;
}

#-------------------------------------------------------------#
# Keep track of the line number from original source.         #
#-------------------------------------------------------------#
{
  ++lineCount;

  if ( lineNo < 1 )
    lineNo = NR;

  # Mark the start of the file by emitting a #line. Note that
  # this code is placed here since the FILENAME variable is 
  # not set until the first line of the file is actually read.
  # That is, it is not valid in the BEGIN section.
  if ( lineCount == 1 )
  {
    if ( substr(FILENAME, 1, 1) != "/" )
      filename = ENVIRON["PWD"] "/" FILENAME;
    else
      filename = FILENAME;
  }
}

#-------------------------------------------------------------#
# Embedded SQL Command                                        #
#-------------------------------------------------------------#
toupper($1)=="EXEC" && toupper($2)=="SQL" {
  # Save the original indentation to make columns line up in the
  # generated code
  match($0, /([[:space:]]*)[Ee]/, items);

  indentStr = items[1];

  # Get and process the SQL statement.
  statement = removeTandemSpecific(getStatement());
  buffer    = statement;

  # Skip the "EXEC SQL" part and get the command type.
  getToken(" ");
  getToken(" ");

  token = toupper(getToken(" "));

  if      ( token == "INCLUDE"  )
    processInclude();
  else if ( token == "BEGIN"    )
    processBegin();
  else if ( token == "END"      )
    processEnd();
  else if ( token == "DECLARE"  )
    processDeclareCursor();
  else if ( token == "OPEN"     )
    processOpenCursor();
  else if ( token == "FETCH"    )
    processFetchCursor();
  else if ( token == "CLOSE"    )
    processCloseCursor();
  else if ( token == "SELECT"   )
    processSelect();
  else if ( token == "INSERT"   )
    processInsert();
  else if ( token == "UPDATE"   )
    processUpdate();
  else if ( token == "DELETE"   )
    processDelete();
  else if ( token == "ROLLBACK" )
    processRollback();
  else if ( token == "COMMIT"   )
    processCommit();
  else if ( token == "CONTROL"  )
    processControl();

  next;
}


#-------------------------------------------------------------#
# Remove any Tandem specific include files.                   #
#-------------------------------------------------------------#
/#include.*(sql\.h|cextdecs\.h)/ {
  code[lineCount] = "//" $0;
  next;
}


#-------------------------------------------------------------#
# Upon detecting the "sqlcode", comment it out as it is       #
# defined in e2mysql.h as a static variable.                  #
#-------------------------------------------------------------#
/(short|int16)[[:space:]]+sqlcode/ {
  code[lineCount] = "// " $0 "  -- Defined in e2mysql.h";
  next;
}


#-------------------------------------------------------------#
# At this point, save any lines not captured above. If the    #
# current line is within a declare section, assure that the   #
# host variables have the static definition so that duplicate #
# external symbols are not detected by the linker.            #
#-------------------------------------------------------------#
{
  if ( inDeclare  )
  {
    if ( ! ($0 ~ /^[[:space:]]*$/ || $0 ~ /^[[:space:]]*(#|\/\/|\/\*)/) )
    {
      statement = getUnbufferedStatement();

      if ( statement ~ /^[[:space:]]*static[[:space:]]/ )
        print statement;
      else
        print "static " statement;
    }

    else
      print $0;

    lineCount =  0;
    lineNo    = -1;
  }

  else
    code[lineCount] = $0
}


#-------------------------------------------------------------#
# Output any lines remaining in the buffer.                   #
#-------------------------------------------------------------#
END {
  outputBufferedCode();
}


#-------------------------------------------------------------#
# Embedded SQL Statements                                     #
#-------------------------------------------------------------#
function processInclude() {
  type = toupper(getToken(" "));

  if ( type == "SQLCA" )
    code[lineCount] = "// #include \"sqlca.h\"";
  else if ( type == "SQLSA" )
    code[lineCount] = "// #include \"sqlsa.h\"";
}


function processBegin(type) {
  type = toupper(getToken(" "));

  if ( type == "WORK" )
  {
    # EXEC SQL BEGIN WORK
    code[esqlLine] = indentStr "esqlBegin();  " code[esqlLine];
  }

  else
  {
    # EXEC SQL BEGIN DECLARE SECTION
    outputBufferedCode();
    inDeclare = 1;
  }
}


function processEnd() {
  inDeclare = 0;
  outputBufferedCode();
}


function processDeclareCursor() {
  cursorName = tolower(getToken(" "));
  sqlBuffer[cursorName] = statement;
  print "// EXEC SQL DECLARE " cursorName;
  declareCursor(cursorName);
  print "//-----------------";
  print " ";
}


function processOpenCursor() {
  cursorName = tolower(getToken(" "));
  code[esqlLine] = indentStr "sqlcode = open_" cursorName "();  " code[esqlLine];

  print "// EXEC SQL OPEN " cursorName;
  openCursor(cursorName, "[Ww][Hh][Ee][Rr][Ee]");
  print "//--------------";
  print " ";
}


function processFetchCursor() {
  cursorName = tolower(getToken(" "));
  code[esqlLine] = indentStr "sqlcode = fetch_" cursorName "();  " code[esqlLine];

  print "// EXEC SQL FETCH " cursorName;
  fetchCursor(cursorName);
  print "//---------------";
  print " ";
}


function processCloseCursor() {
  cursorName = tolower(getToken(" "));
  code[esqlLine] = indentStr "sqlcode = close_" cursorName "();  " code[esqlLine];

  closeCursor(cursorName);
}


function processSelect() {
  cursorName = "tempCursor" uniqCount++;
  sqlBuffer[cursorName] = statement;
  code[esqlLine] = indentStr "sqlcode = select_" cursorName "();  " code[esqlLine];

  print "// EXEC SQL SELECT";
  declareCursor(cursorName);
  openCursor(cursorName, "[Ww][Hh][Ee][Rr][Ee]");
  fetchCursor(cursorName);
  closeCursor(cursorName);
  print "static int16";
  print "select_" cursorName "()";
  print "{";
  print "  sqlcode = open_" cursorName "();";
  print "  if ( sqlcode >= 0 )";
  print "  {";
  print "    sqlcode = fetch_" cursorName "(); ";
  print "    close_" cursorName "(); ";
  print "  }";
  print "  return sqlcode;";
  print "}";
  print "//----------------";
}


function processInsert() {
  cursorName = "tempInsert" uniqCount++;
  sqlBuffer[cursorName] = statement;
  code[esqlLine] = indentStr "sqlcode = open_" cursorName "();  " code[esqlLine];

  print "// EXEC SQL INSERT";
  declareCursor(cursorName);
  openCursor(cursorName, "[Vv][Aa][Ll][Uu][Ee][Ss]");
  print "//----------------";
}


function processUpdate() {
  cursorName = "tempUpdate" uniqCount++;
  sqlBuffer[cursorName] = statement;
  code[esqlLine] = indentStr "sqlcode = open_" cursorName "();  " code[esqlLine];

  print "// EXEC SQL UPDATE";
  declareCursor(cursorName);
  openCursor(cursorName, "[Ww][Hh][Ee][Rr][Ee]");
  print "//----------------";
}


function processDelete() {
  cursorName = "tempDelete" uniqCount++;
  sqlBuffer[cursorName] = statement;
  code[esqlLine] = indentStr "sqlcode = open_" cursorName "();  " code[esqlLine];

  print "// EXEC SQL DELETE";
  declareCursor(cursorName);
  openCursor(cursorName, "[Ww][Hh][Ee][Rr][Ee]");
  print "//----------------";
}


function processRollback() {
  code[esqlLine] = indentStr "esqlRollback();  " code[esqlLine];
}


function processCommit() {
  code[esqlLine] = indentStr "esqlCommit();  " code[esqlLine];
}


function processControl() {
  # EXEC SQL CONTROL ...
  # Do nothing. For the time being, this statement is ignored
  # while processing for MySQL.
}


#-------------------------------------------------------------#
# Utility Functions                                           #
#-------------------------------------------------------------#
function getStatement() {
  stmt     = "";
  done     = 0;
  quoted   = 0;
  esqlLine = lineCount;

  while ( ! done )
  {
    # Construct the statement commented out.
    code[lineCount] = "//" $0;

    # Remove any comments.
    while ( match($0, /(.*)--/, items) )
      $0 = items[1];

    # Find terminating semi-colon for the SQL statement
    len = length($0);

    for ( pos = 1; pos <= len && ! done; ++pos )
    {
      c = substr($0, pos, 1);

      if ( c ~ /[\'\"]/ )
        quoted = 1 - quoted;
      else if ( c == ";" && ! quoted )
        done = 1;
    }

    # Build the statement and grab another line if the end of statement
    # has not been reached. Otherwise, extract the text from the current
    # line up to the terminating semicolon.
    if ( ! done )
    {
      stmt = stmt " " $0;

      getline;

      ++lineCount;

      if ( lineNo < 1 )
        lineNo = lineCount;
    }

    else
      stmt = stmt " " substr($0, 1, pos);
  }

  gsub(/^[[:space:]]+/,             "",  stmt);
  gsub(/[[:space:]]+$/,             "",  stmt);
  gsub(/\[[[:space:]]*/,            "[", stmt);
  gsub(/[[:space:]]*\]/,            "]", stmt);
  gsub(/[[:space:]]*,[[:space:]]*/, ",", stmt);
  gsub(/[[:space:]]*;[[:space:]]*/, ";", stmt);
  gsub(/[[:space:]]+/,              " ", stmt);
  gsub(/[[:space:]]+\(/,            " (", stmt);

  return stmt;
}


function getUnbufferedStatement() {
  stmt     = "";
  done     = 0;
  quoted   = 0;
  esqlLine = lineCount;

  while ( ! done )
  {
    # Find terminating semi-colon for the statement
    len = length($0);

    for ( pos = 1; pos <= len && ! done; ++pos )
    {
      c = substr($0, pos, 1);

      if ( c ~ /[\'\"]/ )
        quoted = 1 - quoted;
      else if ( c == ";" && ! quoted )
        done = 1;
    }

    # Build the statement and grab another line if the end of statement
    # has not been reached.
    stmt = stmt " " $0;

    if ( ! done )
      getline;
  }

  return stmt;
}


function getToken(delimiters) {
  delimiters = "[" delimiters ";]";

  token  = "";
  done   = 0;
  quoted = 0;

  while ( ! done && buffer != "" )
  {
    c      = substr(buffer, 1, 1);
    buffer = substr(buffer, 2   );

    if ( c ~ /[\'\"]/ )
      quoted = 1 - quoted;
    else if ( c ~ delimiters && ! quoted )
      done   = 1;

    if ( ! done )
      token  = token c
  }

  delimiterFound = c;

  return token
}


function declareCursor(cursorName) {
  mySQL = statement;

  if ( match(mySQL, /[Ee][Xx][Ee][Cc][[:space:]]+[Ss][Qq][Ll][[:space:]]+/) )
    mySQL = substr(mySQL, RSTART+RLENGTH);

  if ( match(toupper(mySQL), / (FOR )?(BROWSE|STABLE|REPEATABLE) ACCESS;?/) )
    mySQL = substr(mySQL, 1, RSTART-1) substr(mySQL, RSTART+RLENGTH);

  if ( mySQL ~ /(^|[[:space:]])[Ss][Ee][Ll][Ee][Cc][Tt][[:space:]]/ && match(mySQL, / [Ii][Nn][Tt][Oo] /) )
  {
    pos = RSTART - 1;

    if ( match(mySQL, / [Ff][Rr][Oo][Mm] /) )
      mySQL = substr(mySQL, 1, pos) substr(mySQL, RSTART);
    else
      mySQL = substr(mySQL, 1, pos);
  }

  if ( match(mySQL, /(^|[[:space:]])[Ss][Ee][Ll][Ee][Cc][Tt][[:space:]]/) )
    mySQL = substr(mySQL, RSTART);

  print "static e2mysql " cursorName " = {";
  print "    " quotedString(mySQL);
  print "  , NULL";
  print "  , 0";
  print "};";
}


function openCursor(cursorName, keywords) {
  functionName = "open_" cursorName "()";

  if ( functionName in functionList )
    return;

  functionList[functionName] = 1;

  print "static int16";
  print functionName;
  print "{";

  if ( sqlBuffer[cursorName] ~ /:currentDate/ )
    print "  int64 currentDate = getJulianTimeStamp();";

  if ( sqlBuffer[cursorName] ~ /:intervalDay/ )
    print "  int64 intervalDay = 86400000000;";

  print "  try";
  print "  {";
  print "    if ( ! connectionMade )";
  print "      SQLHelperConnect();";
  print "    if ( "cursorName ".rslt != NULL )";
  print "      mysql_free_result(" cursorName ".rslt);";
  print "    std::string sqlStmt = " cursorName ".stmt;";

  clause = sqlBuffer[cursorName];
  if ( match(sqlBuffer[cursorName], keywords) )
    clause = substr(sqlBuffer[cursorName], RSTART+RLENGTH);

  while ( matchVar(clause) > 0 && clause != "" )
  {
    replTxt = substr(clause, RSTART, RLENGTH);
    hostVar = substr(replTxt, 2);
    clause  = substr(clause, RSTART+RLENGTH);

    print "    SQLBindParmPoly(sqlStmt, \"" replTxt "\", " hostVar ", sizeof(" hostVar "));";
  }

  if ( length(ENVIRON["ATSE_MYSQL_DEBUG"]) > 0 )
    print "fprintf(stderr, \"DEBUG %s(" functionName "): '%s'\\n\", __FILE__, sqlStmt.c_str());";

  print "    if ( mysql_real_query(&mysqlConnection, sqlStmt.c_str(), sqlStmt.length()) )";
  print "      handle_error(\"mysql_real_query\");";
  print "    " cursorName ".rslt  = mysql_store_result(&mysqlConnection);";
  print "    if ( ! " cursorName ".rslt )";
  print "      if ( mysql_field_count(&mysqlConnection) != 0 )";
  print "        handle_error(\"mysql_store_result - no results\");";
  print "    " cursorName ".row  = 0;";
  print "    sqlcode = SQL_SUCCESS;";
  print "  }";
  print "  catch(...)";
  print "  {";
  print "    sqlcode = SQL_ERROR;";
  print "  }";
  print "  return sqlcode;";
  print "}";
}


function fetchCursor(cursorName) {
  functionName = "fetch_" cursorName "()";

  if ( functionName in functionList )
    return;

  functionList[functionName] = 1;

  print "static int16";
  print functionName;
  print "{";
  print "  if ( ! " cursorName ".rslt )";
  print "    return SQL_ERROR;";
  print "  if ( " cursorName ".row >= mysql_num_rows(" cursorName ".rslt) )";
  print "    return SQL_NO_DATA;";
  print "  MYSQL_ROW row = mysql_fetch_row(" cursorName ".rslt);";

  clause = statement;

  if ( match(clause, / [Ii][Nn][Tt][Oo] /) )
    clause = substr(clause, RSTART+RLENGTH);

  if ( match(clause, / [Ff][Rr][Oo][Mm] /) )
    clause = substr(clause, 1, RSTART-1)

  buffer = clause;
  varNo  = 0;

  while ( (hostVar = getToken(" ,")) != "" )
  {
    if ( match(hostVar, /:([[:alpha:]][[:alnum:]._]*)~/, items) )
      var = items[1];
    else
      var = substr(hostVar, 2);

    if ( hostVar ~ /^:/ )
      printf("  SQLBindColPoly(row[%d], %s, sizeof(%s));\n", varNo++, var, var);
  }

  print "  ++" cursorName ".row;";
  print "  return SQL_SUCCESS;";
  print "}";
}


function closeCursor(cursorName) {
  functionName = "close_" cursorName "()";

  if ( functionName in functionList )
    return;

  functionList[functionName] = 1;

  print "// EXEC SQL CLOSE " cursorName;
  print "static int16";
  print functionName;
  print "{";
  print "  mysql_free_result(" cursorName ".rslt);";
  print "  " cursorName ".rslt = NULL;";
  print "  " cursorName ".row  = 0;";
  print "  return SQL_SUCCESS;";
  print "}";
}


function matchVar(sql) {
  return match(sql, /\:[[:alpha:]][[:alnum:]_\.]*(\[[^\]]*\])*/);
}


function quotedString(text) {
  quoted = "\""
  len = length(text)
  for ( i = 1; i <= len; ++i )
  {
    c = substr(text, i, 1);

    if (c == "\"")
      c = "\\\"";

    quoted = quoted c;
  }

  return quoted "\"";
}


function removeTandemSpecific(statement) {
  buffer        = statement;
  output        = "";
  remove        = 0;
  lastDelimiter = " ";

  # Remove "TYPE AS" for dates.
  while (( token = getToken(" ,")) != "" )
  {
    upper = toupper(token);

    if ( remove )
    {
      if ( upper ~ /^(AS|DATE|DATETIME|YEAR|TO|DAY|HOUR|MINUTE|SECOND|FRACTION(\([0-6]\))?)$/ )
        remove = 1;
      else
        remove = 0;
    }

    else if ( upper == "TYPE")
      remove = 1;
    else
      remove = 0;

    if ( ! remove )
    {
      if ( length(output) > 0 )
        output = output lastDelimiter token;
      else
        output = token;
    }

    if ( upper == "CURRENT" )
      remove = 1;

    lastDelimiter = delimiterFound;
  }

  # Convert double quotes allowed by the Himalaya and non-standard to tick marks.
  while ( match(output, /"/) )
    output = substr(output, 1, RSTART-1) "'" substr(output, RSTART+1);

  # Convert the use of the INDICATOR keyword so that it can be easily detected when
  # binding columns.
  while ( match(output, /[[:space:]][Ii][Nn][Dd][Ii][Cc][Aa][Tt][Oo][Rr][[:space:]]/) )
    output = substr(output, 1, RSTART-1) "~" substr(output, RSTART+RLENGTH);

  # Remove Tandem specific built-in functions.
  removeBuiltIn("[Cc][Aa][Ss][Tt][[:space:]]*\\(");
  removeBuiltIn("[Cc][Oo][Nn][Vv][Ee][Rr][Tt][Tt][Ii][Mm][Ee][Ss][Tt][Aa][Mm][Pp][[:space:]]*\\(");
  removeBuiltIn("[Jj][Uu][Ll][Ii][Aa][Nn][Tt][Ii][Mm][Ee][Ss][Tt][Aa][Mm][Pp][[:space:]]*\\(");

  # Define CURRENT and INTERVAL as host varaiables.
  while ( match(output, /([[:space:]]|[[:punct:]])[Cc][Uu][Rr][Rr][Ee][Nn][Tt]([[:space:]]|[[:punct:]])/) )
    output = substr(output, 1, RSTART) ":currentDate" substr(output, RSTART+RLENGTH-1);

  while ( match(output, /([[:space:]]|[[:punct:]])[Ii][Nn][Tt][Ee][Rr][Vv][Aa][Ll][[:space:]]+'1'[[:space:]]+[Dd][Aa][Yy]([[:space:]]|[[:punct:]])/) )
    output = substr(output, 1, RSTART) ":intervalDay" substr(output, RSTART+RLENGTH-1);

  # Fixup Tandem table names.
  if ( match(output, /[Ww][Hh][Ee][Rr][Ee]/) )
  {
    where  = substr(output, RSTART);
    output = substr(output, 1, RSTART-1);
  }

  else
    where = "";

  # Update the table name to remove the leading equal sign requied for the Himalaya defines.
  while ( match(output, /=([[:alpha:]][[:alnum:]]*)/, items) )
    output = substr(output, 1, RSTART-1) "[OWNER]" substr(output, RSTART+1);

  output = output where;

  return output;
}


function removeBuiltIn(pattern) {
  # Remove built-in functions.
  while ( match(output, pattern) )
  {
    pos = RSTART + RLENGTH;
    opn = 1;

    while ( opn && pos <= length(output) )
    {
      if ( substr(output, pos, 1) == "(" )
        ++opn;

      if ( substr(output, pos, 1) == ")" )
        --opn;

      ++pos;
    }

    if ( substr(output, --pos, 1) == ")" )
      output = substr(output, 1, pos - 1) substr(output, pos + 1);

    output = substr(output, 1, RSTART - 1) substr(output, RSTART + RLENGTH);
  }
}


function outputBufferedCode() {
  print "//#line " lineNo " \"" filename "\"";

  for ( i = 1; i <= lineCount; ++i )
    print code[i];

  lineCount =  0;
  lineNo    = -1;
}

