//----------------------------------------------------------------------------
//
//  File:         e2mysql.h
//  Author:       Michael Benzinger
//  Created:      06/26/2002
//  Description:  Helper functions for ESQL -> MySQL preprocessor for use only
//                by ATSE Hybrid projects.
//
//                Based on original work by Alan Walker, Sabre Labs.
//
// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
// USA.
//
//----------------------------------------------------------------------------


#ifndef E2MYSQL_H
#define E2MYSQL_H

#include <stdio.h>
#include <stdint.h>
#include <limits.h>
#include <stdexcept>
#include <mysql.h>

#include <sys/time.h>


// Define MySQL variables.
#ifdef DEFINE_MYSQL_VARIABLES
bool  connectionMade = false;
MYSQL mysqlConnection;
#else
extern bool  connectionMade;
extern MYSQL mysqlConnection;
#endif


static int16_t  sqlcode;


// Define structure to hold SQL statement information.
struct e2mysql
{
  const char* stmt;
  MYSQL_RES*  rslt;
  uint32_t    row;
};


// SQLCA and SQLSA definitions for compatibility with Tandem Himalaya
static struct sqlca_struct
{
  int sqlcode;
  int fileCode;
  int misc;
} sqlca;


#define SQLCAFSCODE(_parm) (sqlca.fileCode)
#define SQLCATOBUFFER(x,y,z) snprintf(y, z, "%hd", *x)
#define TESTSQLCODE(_parm)


#define SQL_SUCCESS    0
#define SQL_ERROR     -1
#define SQL_NO_DATA  100


//#################################################################
//#####        Connection and Error Handling Functions        #####
//#################################################################
int32_t
SQLHelperConnect(void);

void
SQLHelperDisconnect(void);

void
handle_error(const char *msg);


//#################################################################
//#####          Tandem Specific Function Emulations          #####
//#################################################################
inline int32_t
BEGINTRANSACTION(long* /*txnID*/)
{
  try
  {
    if ( mysql_real_query(&mysqlConnection, "BEGIN", 5) )
      handle_error("BEGIN");
    sqlcode = SQL_SUCCESS;
  }
  catch(...)
  {
    sqlcode = SQL_ERROR;
  }
  return sqlcode;
}

inline int32_t
ENDTRANSACTION()
{
  try
  {
    if ( mysql_real_query(&mysqlConnection, "COMMIT", 6) )
      handle_error("COMMIT");
    sqlcode = SQL_SUCCESS;
  }
  catch(...)
  {
    sqlcode = SQL_ERROR;
  }
  return sqlcode;
}

inline int32_t
RESUMETRANSACTION(long /*txnID*/)
{
  return SQL_SUCCESS;
}

inline int32_t
ABORTTRANSACTION()
{
  try
  {
    if ( mysql_real_query(&mysqlConnection, "ROLLBACK", 8) )
      handle_error("ROLLBACK");
    sqlcode = SQL_SUCCESS;
  }
  catch(...)
  {
    sqlcode = SQL_ERROR;
  }
  return sqlcode;
}

inline int64_t
getJulianTimeStamp()
{
  static const int64_t microsecondsToUNIXEpoch = 210866760000000000;

  struct timeval timeOfDay;
  gettimeofday(&timeOfDay, NULL);

  struct tm* pt = localtime(&timeOfDay.tv_sec);

  return (static_cast<int64_t>(mktime(pt)) * 1000000LL + timeOfDay.tv_usec) + microsecondsToUNIXEpoch;
}

//#################################################################
//#####          Bind Host Parameters for the Query           #####
//#################################################################
inline int32_t
SQLBindParmPoly(std::string& stmt, const char* text, const char parm, uint16_t /*size*/)
{
  std::string::size_type pos = stmt.find(text);
  if ( pos != std::string::npos )
  {
    char buffer[4];
    snprintf(buffer, 4, "'%c'", parm);
    stmt.replace(pos, strlen(text), buffer);
  }
  return SQL_SUCCESS;
}

inline int32_t
SQLBindParmPoly(std::string& stmt, const char* text, const char* parm, uint16_t size)
{
  std::string::size_type pos = stmt.find(text);
  if ( pos != std::string::npos )
  {
    --size;
    char buffer[4096];
    snprintf(buffer, 4096, "'%-*.*s'", size, size, parm);
    stmt.replace(pos, strlen(text), buffer);
  }
  return SQL_SUCCESS;
}

inline int32_t
SQLBindParmPoly(std::string& stmt, const char* text, const int16_t parm, uint16_t /*size*/)
{
  std::string::size_type pos = stmt.find(text);
  if ( pos != std::string::npos )
  {
    char buffer[7];
    snprintf(buffer, 7, "%hd", parm);
    stmt.replace(pos, strlen(text), buffer);
  }
  return SQL_SUCCESS;
}

inline int32_t
SQLBindParmPoly(std::string& stmt, const char* text, const uint16_t parm, uint16_t /*size*/)
{
  std::string::size_type pos = stmt.find(text);
  if ( pos != std::string::npos )
  {
    char buffer[6];
    snprintf(buffer, 6, "%hu", parm);
    stmt.replace(pos, strlen(text), buffer);
  }
  return SQL_SUCCESS;
}

inline int32_t
SQLBindParmPoly(std::string& stmt, const char* text, const int32_t parm, uint16_t /*size*/)
{
  std::string::size_type pos = stmt.find(text);
  if ( pos != std::string::npos )
  {
    char buffer[12];
    snprintf(buffer, 12, "%d", parm);
    stmt.replace(pos, strlen(text), buffer);
  }
  return SQL_SUCCESS;
}

inline int32_t
SQLBindParmPoly(std::string& stmt, const char* text, const uint32_t parm, uint16_t /*size*/)
{
  std::string::size_type pos = stmt.find(text);
  if ( pos != std::string::npos )
  {
    char buffer[11];
    snprintf(buffer, 11, "%u", parm);
    stmt.replace(pos, strlen(text), buffer);
  }
  return SQL_SUCCESS;
}

inline int32_t
SQLBindParmPoly(std::string& stmt, const char* text, const int64_t parm, uint16_t /*size*/)
{
  std::string::size_type pos = stmt.find(text);
  if ( pos != std::string::npos )
  {
    char buffer[21];
    snprintf(buffer, 21, "%lld", parm);
    stmt.replace(pos, strlen(text), buffer);
  }
  return SQL_SUCCESS;
}

inline int32_t
SQLBindParmPoly(std::string& stmt, const char* text, const float parm, uint16_t /*size*/)
{
  std::string::size_type pos = stmt.find(text);
  if ( pos != std::string::npos )
  {
    char buffer[21];
    snprintf(buffer, 21, "%f", parm);
    stmt.replace(pos, strlen(text), buffer);
  }
  return SQL_SUCCESS;
}

inline int32_t
SQLBindParmPoly(std::string& stmt, const char* text, const double parm, uint16_t /*size*/)
{
  std::string::size_type pos = stmt.find(text);
  if ( pos != std::string::npos )
  {
    char buffer[21];
    snprintf(buffer, 21, "%f", parm);
    stmt.replace(pos, strlen(text), buffer);
  }
  return SQL_SUCCESS;
}


//#################################################################
//#####          Bind Host Variables to Accept Data           #####
//#################################################################
inline int32_t
SQLBindColPoly(const char* value, char& parm, uint16_t /*size*/)
{
  parm = value ? value[0] : ' ';
  return SQL_SUCCESS;
}

inline int32_t
SQLBindColPoly(const char* value, char* parm, uint16_t size)
{
  if ( value )
  {
    --size;                       // Decrement the size by one to reflect the null terminator.
    const char* s = value;        // Get pointer to the start of the source string.
    char* p =  parm;              // Get pointer to the start of the target string.
    char* e = &parm[size];        // Get pointer to the end   of the target string.
    while ( *s ) *p++ = *s++;     // Assign each byte from the source to the target until a null.
    while ( p != e ) *p++ = ' ';  // Pad the string with blanks since MySQL removes trailing blanks.
    parm[size] = '\0';            // Terminte the string with a null.
  }
  else
    parm[0] = '\0';
  return SQL_SUCCESS;
}

inline int32_t
SQLBindColPoly(const char* value, int16_t& parm, uint16_t /*size*/)
{
  parm = value ? atoi(value) : 0;
  return SQL_SUCCESS;
}

inline int32_t
SQLBindColPoly(const char* value, uint16_t& parm, uint16_t /*size*/)
{
  parm = value ? atoi(value) : 0;
  return SQL_SUCCESS;
}

inline int32_t
SQLBindColPoly(const char* value, int32_t& parm, uint16_t /*size*/)
{
  parm = value ? atoi(value) : 0;
  return SQL_SUCCESS;
}

inline int32_t
SQLBindColPoly(const char* value, uint32_t& parm, uint16_t /*size*/)
{
  parm = value ? atoi(value) : 0;
  return SQL_SUCCESS;
}

inline int32_t
SQLBindColPoly(const char* value, int64_t& parm, uint16_t /*size*/)
{
#if ULONG_MAX == 0xFFFFFFFFul        // 32-bit longs
  parm = value ? atoll(value) : 0;
#else                                // 64-bit longs
  parm = value ? atol (value) : 0;
#endif
  return SQL_SUCCESS;
}

inline int32_t
SQLBindColPoly(const char* value, float& parm, uint16_t /*size*/)
{
  parm = value ? atof(value) : 0;
  return SQL_SUCCESS;
}

inline int32_t
SQLBindColPoly(const char* value, double& parm, uint16_t /*size*/)
{
  parm = value ? atof(value) : 0;
  return SQL_SUCCESS;
}


//#################################################################
//#####             Rollback and Commit Functions             #####
//#################################################################
inline void
esqlBegin()
{
  BEGINTRANSACTION(NULL);
}


inline void
esqlRollback()
{
  ABORTTRANSACTION();
}


inline void
esqlCommit()
{
  ENDTRANSACTION();
}

#endif  // E2MYSQL_H


