#-------------------------------------------------------------------------#
# e2odbc.awk                                                              #
#                                                                         #
# Simple ESQL precompiler to convert embedded SQL for use                 #
# with an ODBC data source.                                               #
#                                                                         #
# Michael Benzinger, Sabre, June 2002                                     #
#                                                                         #
# This script was adapted from original work done by Alan                 #
# Walker, Sabre Labs.                                                     #
#                                                                         #
# This is free software; you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation; either version 2 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# This software is distributed in the hope that it will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this software; if not, write to the Free Software            #
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,  #
# USA.                                                                    #
#-------------------------------------------------------------------------#
BEGIN {
  print "// Code precompiled by e2odbc.awk";
  print "#include \"e2odbc.h\"";
  print "#include \"IniParams.h\"";
  print " ";
  print "static const char* tableOwnerTemplate = \"[OWNER]\";";

  buffer    = "";
  indentStr = "    ";
  lineCount =  0;
  lineNo    = -1;
  uniqCount =  1;
  ifCount   =  0;
  tandem    =  0;
  inDeclare =  0;
}

#-------------------------------------------------------------#
# Keep track of the line number from original source.         #
#-------------------------------------------------------------#
{
  ++lineCount;

  if ( lineNo < 1 )
    lineNo = NR;

  # Mark the start of the file by emitting a #line. Note that
  # this code is placed here since the FILENAME variable is 
  # not set until the first line of the file is actually read.
  # That is, it is not valid in the BEGIN section.
  if ( lineCount == 1 )
  {
    if ( substr(FILENAME, 1, 1) != "/" )
      filename = ENVIRON["PWD"] "/" FILENAME;
    else
      filename = FILENAME;
  }
}

#-------------------------------------------------------------#
# Embedded SQL Command                                        #
#-------------------------------------------------------------#
toupper($1)=="EXEC" && toupper($2)=="SQL" {
  # Save the original indentation to make columns line up in the
  # generated code
  match($0, /([[:space:]]*)[Ee]/, items);

  indentStr = items[1];

  # Get and process the SQL statement.
  statement = removeTandemSpecific(getStatement());
  buffer    = statement;

  # Skip the "EXEC SQL" part and get the command type.
  getToken(" ");
  getToken(" ");

  token = toupper(getToken(" "));

  if      ( token == "INCLUDE"  )
    processInclude();
  else if ( token == "BEGIN"    )
    processBegin();
  else if ( token == "END"      )
    processEnd();
  else if ( token == "DECLARE"  )
    processDeclareCursor();
  else if ( token == "OPEN"     )
    processOpenCursor();
  else if ( token == "FETCH"    )
    processFetchCursor();
  else if ( token == "CLOSE"    )
    processCloseCursor();
  else if ( token == "SELECT"   )
    processSelect();
  else if ( token == "INSERT"   )
    processInsert();
  else if ( token == "UPDATE"   )
    processUpdate();
  else if ( token == "DELETE"   )
    processDelete();
  else if ( token == "ROLLBACK" )
    processRollback();
  else if ( token == "COMMIT"   )
    processCommit();
  else if ( token == "CONTROL"  )
    processControl();

  next;
}


#-------------------------------------------------------------#
# Remove any Tandem specific include files.                   #
#-------------------------------------------------------------#
/#include.*(sql\.h|cextdecs\.h)/ {
  code[lineCount] = "//" $0;
  next;
}


#-------------------------------------------------------------#
# Upon detecting the "sqlcode", comment it out as it is       #
# defined in e2odbc.h as a static variable.                   #
#-------------------------------------------------------------#
/(short|int16)[[:space:]]+sqlcode/ {
  code[lineCount] = "// " $0 "  -- Defined in e2odbc.h";
  next;
}


#-------------------------------------------------------------#
# At this point, save any lines not captured above. If the    #
# current line is within a declare section, assure that the   #
# host variables have the static definition so that duplicate #
# external symbols are not detected by the linker.            #
#-------------------------------------------------------------#
{
  if ( inDeclare  )
  {
    if ( ! ($0 ~ /^[[:space:]]*$/ || $0 ~ /^[[:space:]]*(#|\/\/|\/\*)/) )
    {
      statement = getUnbufferedStatement();

      if ( statement ~ /^[[:space:]]*static[[:space:]]/ )
        print statement;
      else
        print "static " statement;
    }

    else
      print $0;

    lineCount =  0;
    lineNo    = -1;
  }

  else
    code[lineCount] = $0
}


#-------------------------------------------------------------#
# Output any lines remaining in the buffer.                   #
#-------------------------------------------------------------#
END {
  outputBufferedCode();
}


#-------------------------------------------------------------#
# Embedded SQL Statements                                     #
#-------------------------------------------------------------#
function processInclude() {
  type = toupper(getToken(" "));

  if ( type == "SQLCA" )
    code[lineCount] = "// #include \"sqlca.h\"";
  else if ( type == "SQLSA" )
    code[lineCount] = "// #include \"sqlsa.h\"";
}


function processBegin(type) {
  type = toupper(getToken(" "));

  if ( type == "WORK" )
  {
    # EXEC SQL BEGIN WORK
    # Nothing to do other than to comment out the statement.
  }

  else
  {
    # EXEC SQL BEGIN DECLARE SECTION
    outputBufferedCode();
    inDeclare = 1;
  }
}


function processEnd() {
  inDeclare = 0;
  outputBufferedCode();
}


function processDeclareCursor() {
  cursorName = tolower(getToken(" "));
  sqlBuffer[cursorName] = statement;
  print "// EXEC SQL DECLARE " cursorName;
  declareCursor(cursorName);
  print "//-----------------";
  print " ";
}


function processOpenCursor() {
  cursorName = tolower(getToken(" "));
  code[esqlLine] = indentStr "sqlcode = open_" cursorName "();  " code[esqlLine];

  print "// EXEC SQL OPEN " cursorName;
  openCursor(cursorName, "[Ww][Hh][Ee][Rr][Ee]");
  print "//--------------";
  print " ";
}


function processFetchCursor() {
  cursorName = tolower(getToken(" "));
  code[esqlLine] = indentStr "sqlcode = fetch_" cursorName "();  " code[esqlLine];

  print "// EXEC SQL FETCH " cursorName;
  fetchCursor(cursorName);
  print "//---------------";
  print " ";
}


function processCloseCursor() {
  cursorName = tolower(getToken(" "));
  code[esqlLine] = indentStr "sqlcode = close_" cursorName "();  " code[esqlLine];

  closeCursor(cursorName);
}


function processSelect() {
  cursorName = "tempCursor" uniqCount++;
  sqlBuffer[cursorName] = statement;
  code[esqlLine] = indentStr "sqlcode = select_" cursorName "();  " code[esqlLine];

  print "// EXEC SQL SELECT";
  declareCursor(cursorName);
  openCursor(cursorName, "[Ww][Hh][Ee][Rr][Ee]");
  fetchCursor(cursorName);
  closeCursor(cursorName);
  print "static int16";
  print "select_" cursorName "()";
  print "{";
  print "  sqlcode = open_" cursorName "();";
  print "  if ( sqlcode >= 0 )";
  print "  {";
  print "    sqlcode = fetch_" cursorName "(); ";
  print "    close_" cursorName "(); ";
  print "  }";
  print "  return sqlcode;";
  print "}";
  print "//----------------";
}


function processInsert() {
  cursorName = "tempInsert" uniqCount++;
  sqlBuffer[cursorName] = statement;
  code[esqlLine] = indentStr "sqlcode = open_" cursorName "();  " code[esqlLine];

  print "// EXEC SQL INSERT";
  declareCursor(cursorName);
  openCursor(cursorName, "[Vv][Aa][Ll][Uu][Ee][Ss]");
  print "//----------------";
}


function processUpdate() {
  cursorName = "tempUpdate" uniqCount++;
  sqlBuffer[cursorName] = statement;
  code[esqlLine] = indentStr "sqlcode = open_" cursorName "();  " code[esqlLine];

  print "// EXEC SQL UPDATE";
  declareCursor(cursorName);
  openCursor(cursorName, "[Ww][Hh][Ee][Rr][Ee]");
  print "//----------------";
}


function processDelete() {
  cursorName = "tempDelete" uniqCount++;
  sqlBuffer[cursorName] = statement;
  code[esqlLine] = indentStr "sqlcode = open_" cursorName "();  " code[esqlLine];

  print "// EXEC SQL DELETE";
  declareCursor(cursorName);
  openCursor(cursorName, "[Ww][Hh][Ee][Rr][Ee]");
  print "//----------------";
}


function processRollback() {
  code[esqlLine] = indentStr "esqlRollback();  " code[esqlLine];
}


function processCommit() {
  code[esqlLine] = indentStr "esqlCommit();  " code[esqlLine];
}


function processControl() {
  # EXEC SQL CONTROL ...
  # Do nothing. For the time being, this statement is ignored
  # while processing for ODBC.
}


#-------------------------------------------------------------#
# Utility Functions                                           #
#-------------------------------------------------------------#
function getStatement() {
  stmt     = "";
  done     = 0;
  quoted   = 0;
  esqlLine = lineCount;

  while ( ! done )
  {
    # Construct the statement commented out.
    code[lineCount] = "//" $0;

    # Remove any comments.
    while ( match($0, /(.*)--/, items) )
      $0 = items[1];

    # Find terminating semi-colon for the SQL statement
    len = length($0);

    for ( pos = 1; pos <= len && ! done; ++pos )
    {
      c = substr($0, pos, 1);

      if ( c ~ /[\'\"]/ )
        quoted = 1 - quoted;
      else if ( c == ";" && ! quoted )
        done = 1;
    }

    # Build the statement and grab another line if the end of statement
    # has not been reached. Otherwise, extract the text from the current
    # line up to the terminating semicolon.
    if ( ! done )
    {
      stmt = stmt " " $0;

      getline;

      ++lineCount;

      if ( lineNo < 1 )
        lineNo = lineCount;
    }

    else
      stmt = stmt " " substr($0, 1, pos);
  }

  gsub(/^[[:space:]]+/,             "",  stmt);
  gsub(/[[:space:]]+$/,             "",  stmt);
  gsub(/\[[[:space:]]*/,            "[", stmt);
  gsub(/[[:space:]]*\]/,            "]", stmt);
  gsub(/[[:space:]]*,[[:space:]]*/, ",", stmt);
  gsub(/[[:space:]]*;[[:space:]]*/, ";", stmt);
  gsub(/[[:space:]]+/,              " ", stmt);

  return stmt;
}


function getUnbufferedStatement() {
  stmt     = "";
  done     = 0;
  quoted   = 0;
  esqlLine = lineCount;

  while ( ! done )
  {
    # Find terminating semi-colon for the statement
    len = length($0);

    for ( pos = 1; pos <= len && ! done; ++pos )
    {
      c = substr($0, pos, 1);

      if ( c ~ /[\'\"]/ )
        quoted = 1 - quoted;
      else if ( c == ";" && ! quoted )
        done = 1;
    }

    # Build the statement and grab another line if the end of statement
    # has not been reached.
    stmt = stmt " " $0;

    if ( ! done )
      getline;
  }

  return stmt;
}


function getToken(delimiters) {
  delimiters = "[" delimiters ";]";

  token  = "";
  done   = 0;
  quoted = 0;

  while ( ! done && buffer != "" )
  {
    c      = substr(buffer, 1, 1);
    buffer = substr(buffer, 2   );

    if ( c ~ /[\'\"]/ )
      quoted = 1 - quoted;
    else if ( c ~ delimiters && ! quoted )
      done   = 1;

    if ( ! done )
      token  = token c
  }

  delimiterFound = c;

  return token
}


function declareCursor(cursorName) {
  # Replace the host variables with question marks per ODBC specification.
  odbcSQL = statement;

  if ( odbcSQL ~ /[[:space:]][Ss][Ee][Ll][Ee][Cc][Tt][[:space:]]/ && match(odbcSQL, / [Ii][Nn][Tt][Oo] /) )
  {
    pos = RSTART - 1;

    if ( match(odbcSQL, / [Ff][Rr][Oo][Mm] /) )
      odbcSQL = substr(odbcSQL, 1, pos) substr(odbcSQL, RSTART);
    else
      odbcSQL = substr(odbcSQL, 1, pos);
  }

  while ( matchVar(odbcSQL) > 0 )
    odbcSQL = substr(odbcSQL, 1, RSTART-1) "?" substr(odbcSQL, RSTART+RLENGTH);

  if ( match(odbcSQL, /[[:space:]][Ss][Ee][Ll][Ee][Cc][Tt][[:space:]]/) )
  {
    odbcSQL = substr(odbcSQL, RSTART);

    sub(/,[Ll][Ee][Vv][Ee][Ll],/, ",xlevel,", odbcSQL);
  }

  if ( match(odbcSQL, /[Ee][Xx][Ee][Cc][[:space:]]+[Ss][Qq][Ll][[:space:]]+/) )
    odbcSQL = substr(odbcSQL, RSTART+RLENGTH);

  if ( match(toupper(odbcSQL), / (FOR )?(BROWSE|STABLE|REPEATABLE) ACCESS;?/) )
    odbcSQL = substr(odbcSQL, 1, RSTART-1) substr(odbcSQL, RSTART+RLENGTH);

  print "static e2odbc " cursorName " = {";
  print "    " quotedString(odbcSQL);
  print "  , false";
  print "  , SQL_NULL_HSTMT";
  print "};";
}


function openCursor(cursorName, keywords) {
  functionName = "open_" cursorName "()";

  if ( functionName in functionList )
    return;

  functionList[functionName] = 1;

  print "static int16";
  print functionName;
  print "{";

  if ( sqlBuffer[cursorName] ~ /:currentDate/ )
    print "  int64 currentDate = getJulianTimeStamp();";

  if ( sqlBuffer[cursorName] ~ /:intervalDay/ )
    print "  int64 intervalDay = 86400000000;";

  print "  try";
  print "  {";
  print "    if (" cursorName ".stmt == SQL_NULL_HSTMT)";
  print "    {";
  print "      if ( odbcConnection == 0 )";
  print "        SQLHelperConnect();";
  print "      sqlcode = SQLAllocStmt(odbcConnection, &" cursorName ".stmt);";
  print "      if ( sqlcode == SQL_ERROR || sqlcode == SQL_INVALID_HANDLE )";
  print "        handle_error(odbcEnvironment, odbcConnection, " cursorName ".stmt, sqlcode, \"SQLAllocStmt\");";
  print "      atse::IniParams& iniHandler = atse::IniParams::getGlobal();";
  print "      std::string sqlStatementText(" cursorName ".text);";
  print "      std::string tableOwnerName;";
  print "      if ( iniHandler.find(\"Hybrid\", \"TableOwnerName\", tableOwnerName) )";
  print "        tableOwnerName += \".\";";
  print "      std::string::size_type pos;";
  print "      while ( (pos = sqlStatementText.find(tableOwnerTemplate)) != std::string::npos )";
  print "        sqlStatementText.replace(pos, strlen(tableOwnerTemplate), tableOwnerName);";
  print "      sqlcode = SQLPrepare(" cursorName ".stmt, (SQLCHAR*) sqlStatementText.c_str(), SQL_NTS);";
  print "      if ( sqlcode != SQL_SUCCESS && sqlcode != SQL_SUCCESS_WITH_INFO )";
  print "        handle_error(odbcEnvironment, odbcConnection, " cursorName ".stmt, sqlcode, \"SQLPrepare\");";
  bindParms(cursorName, keywords);
  print "    }";
  print "    sqlcode = SQLExecute(" cursorName ".stmt);";
  print "    if ( sqlcode != SQL_SUCCESS && sqlcode != SQL_SUCCESS_WITH_INFO )";
  print "      handle_error(odbcEnvironment, odbcConnection, " cursorName ".stmt, sqlcode, \"SQLExecute\");";
  print "  }";
  print "  catch(...)";
  print "  {";
  print "  }";
  print "  return sqlcode;";
  print "}";
}


function bindParms(cursorName, keywords) {
  clause = sqlBuffer[cursorName];
  if ( match(sqlBuffer[cursorName], keywords) )
    clause = substr(sqlBuffer[cursorName], RSTART+RLENGTH);

  parmNo = 1;

  while ( matchVar(clause) > 0 && clause != "" && parmNo < 256)
  {
    hostVar = substr(clause, RSTART+1, RLENGTH-1);
    clause  = substr(clause, RSTART+RLENGTH);

    print "      SQLBindParmPoly(" cursorName ".stmt, " parmNo++ "," hostVar ", sizeof(" hostVar "));";
  }
}


function fetchCursor(cursorName) {
  functionName = "fetch_" cursorName "()";

  if ( functionName in functionList )
    return;

  functionList[functionName] = 1;

  print "static int16";
  print functionName;
  print "{";

  clause = statement;

  if ( match(clause, / [Ii][Nn][Tt][Oo] /) )
    clause = substr(clause, RSTART+RLENGTH);

  if ( match(clause, / [Ff][Rr][Oo][Mm] /) )
    clause = substr(clause, 1, RSTART-1)

  buffer = clause;
  indNo  = 1;

  while ( (hostVar = getToken(" ,")) != "" )
    if ( hostVar ~ /~/ )
      print "  static SQLLEN esqlIndicator_" indNo++ ";";

  buffer = clause;
  varNo  = 1;
  indNo  = 1;

  print "  try";
  print "  {";
  print "    if ( ! " cursorName ".bind )";
  print "    {";

  while ( (hostVar = getToken(" ,")) != "" )
  {
    if ( match(hostVar, /:([[:alpha:]][[:alnum:]._]*)~/, items) )
    {
      var = items[1];
      ind = "esqlIndicator_" indNo++;
    }

    else
    {
      var = substr(hostVar, 2);
      ind = "dummyNullInd";
    }

    if ( hostVar ~ /^:/ )
      print "      SQLBindColPoly(" cursorName ".stmt, " varNo++ ", " var ", sizeof(" var "), &" ind ");";
  }

  print "      " cursorName ".bind = true;";
  print "    }";
  print "    sqlcode = SQLFetch(" cursorName ".stmt);"
  print "    if ( sqlcode == SQL_ERROR || sqlcode == SQL_INVALID_HANDLE )";
  print "      handle_error(odbcEnvironment, odbcConnection, " cursorName ".stmt, sqlcode, \"SQLFetch\");";

  buffer = clause;
  indNo  = 1;

  while ( (hostVar = getToken(" ,")) != "" )
    if ( match(hostVar, /~:([[:alpha:]][[:alnum:]._]*)/, items) )
      print "    " items[1] " = esqlIndicator_" indNo++ ";";

  print "  }";
  print "  catch(...)";
  print "  {";
  print "  }";
  print "  return sqlcode;";
  print "}";
}


function closeCursor(cursorName) {
  functionName = "close_" cursorName "()";

  if ( functionName in functionList )
    return;

  functionList[functionName] = 1;

  print "// EXEC SQL CLOSE " cursorName;
  print "static int16";
  print functionName;
  print "{";
  print "  try";
  print "  {";
  print "    sqlcode = SQLFreeStmt(" cursorName ".stmt, SQL_CLOSE);";
  print "    if ( sqlcode == SQL_ERROR || sqlcode == SQL_INVALID_HANDLE )";
  print "      handle_error(odbcEnvironment, odbcConnection, " cursorName ".stmt, sqlcode, \"SQLFreeStmt\");";
  print "  }";
  print "  catch(...)";
  print "  {";
  print "  }";
  print "  return sqlcode;";
  print "}";
}


function matchVar(sql) {
  return match(sql, /\:[[:alpha:]][[:alnum:]_\.]*(\[[^\]]*\])*/);
}


function quotedString(text) {
  quoted = "\""
  len = length(text)
  for ( i = 1; i <= len; ++i )
  {
    c = substr(text, i, 1);

    if (c == "\"")
      c = "\\\"";

    quoted = quoted c;
  }

  return quoted "\"";
}


function removeTandemSpecific(statement) {
  buffer        = statement;
  output        = "";
  remove        = 0;
  lastDelimiter = " ";

  # Remove "TYPE AS" for dates.
  while (( token = getToken(" ,")) != "" )
  {
    upper = toupper(token);

    if ( remove )
    {
      if ( upper ~ /^(AS|DATE|DATETIME|YEAR|TO|DAY|HOUR|MINUTE|SECOND|FRACTION(\([0-6]\))?)$/ )
        remove = 1;
      else
        remove = 0;
    }

    else if ( upper == "TYPE")
      remove = 1;
    else
      remove = 0;

    if ( ! remove )
    {
      if ( length(output) > 0 )
        output = output lastDelimiter token;
      else
        output = token;
    }

    if ( upper == "CURRENT" )
      remove = 1;

    lastDelimiter = delimiterFound;
  }

  # Convert double quotes allowed by the Himalaya and non-standard to tick marks.
  while ( match(output, /"/) )
    output = substr(output, 1, RSTART-1) "'" substr(output, RSTART+1);

  # Convert the use of the INDICATOR keyword so that it can be easily detected when
  # binding columns.
  while ( match(output, /[[:space:]][Ii][Nn][Dd][Ii][Cc][Aa][Tt][Oo][Rr][[:space:]]/) )
    output = substr(output, 1, RSTART-1) "~" substr(output, RSTART+RLENGTH);

  # Remove Tandem specific built-in functions.
  removeBuiltIn("[Cc][Aa][Ss][Tt][[:space:]]*\\(");
  removeBuiltIn("[Cc][Oo][Nn][Vv][Ee][Rr][Tt][Tt][Ii][Mm][Ee][Ss][Tt][Aa][Mm][Pp][[:space:]]*\\(");
  removeBuiltIn("[Jj][Uu][Ll][Ii][Aa][Nn][Tt][Ii][Mm][Ee][Ss][Tt][Aa][Mm][Pp][[:space:]]*\\(");

  # Define CURRENT and INTERVAL as host varaiables.
  while ( match(output, /([[:space:]]|[[:punct:]])[Cc][Uu][Rr][Rr][Ee][Nn][Tt]([[:space:]]|[[:punct:]])/) )
    output = substr(output, 1, RSTART) ":currentDate" substr(output, RSTART+RLENGTH-1);

  while ( match(output, /([[:space:]]|[[:punct:]])[Ii][Nn][Tt][Ee][Rr][Vv][Aa][Ll][[:space:]]+'1'[[:space:]]+[Dd][Aa][Yy]([[:space:]]|[[:punct:]])/) )
    output = substr(output, 1, RSTART) ":intervalDay" substr(output, RSTART+RLENGTH-1);

  # Fixup Tandem table names.
  if ( match(output, /[Ww][Hh][Ee][Rr][Ee]/) )
  {
    where  = substr(output, RSTART);
    output = substr(output, 1, RSTART-1);
  }

  else
    where = "";

  while ( match(output, /=[[:alpha:]][[:alnum:]]*/) )
    output = substr(output, 1, RSTART-1) "[OWNER]" substr(output, RSTART+1);

  output = output where;

  return output;
}


function removeBuiltIn(pattern) {
  # Remove built-in functions.
  while ( match(output, pattern) )
  {
    pos = RSTART + RLENGTH;
    opn = 1;

    while ( opn && pos <= length(output) )
    {
      if ( substr(output, pos, 1) == "(" )
        ++opn;

      if ( substr(output, pos, 1) == ")" )
        --opn;

      ++pos;
    }

    if ( substr(output, --pos, 1) == ")" )
      output = substr(output, 1, pos - 1) substr(output, pos + 1);

    output = substr(output, 1, RSTART - 1) substr(output, RSTART + RLENGTH);
  }
}


function outputBufferedCode() {
  print "#line " lineNo " \"" filename "\"";

  for ( i = 1; i <= lineCount; ++i )
    print code[i];

  lineCount =  0;
  lineNo    = -1;
}

